@@ -20,6 +20,7 @@
 #include <sys/uio.h>
 #endif /* __FreeBSD__ */
 #include <signal.h>
+#include <gdk/gdk.h>
 
 #ifdef USE_HAL
 #include <libhal.h>
@@ -334,7 +335,7 @@ growisofs_stderr_read (GIOChannel   *source,
 
 	/* TODO: Handle errors */
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
-		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
+		g_string_append (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "unsupported MMC profile") != NULL || (strstr (line, "already carries isofs") != NULL && strstr (line, "FATAL:") != NULL)) {
 			/* This is not supposed to happen since we checked for the cd
 			   type before starting, but we try to handle it anyway, since mmc
@@ -397,7 +398,7 @@ cdrecord_stderr_read (GIOChannel   *source,
 
 	/* TODO: Handle errors */
 	if (status == G_IO_STATUS_NORMAL && !cdrecorder->priv->expect_cdrecord_to_die) {
-		g_string_prepend (cdrecorder->priv->cdr_stderr, line);
+		g_string_append (cdrecorder->priv->cdr_stderr, line);
 		if (strstr (line, "No disk / Wrong disk!") != NULL) {
 			/* This is not supposed to happen since we checked for the cd
 			   before starting, but we try to handle it anyway, since mmc
@@ -486,12 +487,183 @@ media_type_matches (CDMediaType type)
 	return FALSE;
 }
 
+typedef struct {
+	gboolean timeout;
+	gboolean unmount_ok;
+	guint timeout_tag;
+	GMainLoop *loop;
+	char *device;
+	const char *command;
+} UnmountData;
+
+static const char *umount_known_locations [] = {
+	"/sbin/umount", "/bin/umount",
+	"/usr/sbin/umount", "/usr/bin/umount",
+	NULL
+};
+
+static void
+free_unmount_data (UnmountData *unmount_data)
+{
+	g_free (unmount_data->device);
+	g_free (unmount_data);
+}
+
+static gboolean
+unmount_done (gpointer data)
+{
+	UnmountData *unmount_data;
+	unmount_data = data;
+	
+	if (unmount_data->timeout_tag != 0) {
+		g_source_remove (unmount_data->timeout_tag);
+	}
+
+	if (unmount_data->loop != NULL &&
+	    g_main_loop_is_running (unmount_data->loop)) {
+		g_main_loop_quit (unmount_data->loop);
+	}
+	
+	if (unmount_data->timeout) {
+		/* We timed out, so unmount_data wasn't freed
+		   at mainloop exit. */
+		free_unmount_data (unmount_data);
+	}
+	
+	return FALSE;
+}
+
+static gboolean
+unmount_timeout (gpointer data)
+{
+	UnmountData *unmount_data;
+	unmount_data = data;
+
+	/* We're sure, the callback hasn't been run, so just say
+	   we were interrupted and return from the mainloop */
+	
+	unmount_data->unmount_ok = FALSE;
+	unmount_data->timeout_tag = 0;
+	unmount_data->timeout = TRUE;
+	
+	if (g_main_loop_is_running (unmount_data->loop)) {
+		g_main_loop_quit (unmount_data->loop);
+	}
+	
+	return FALSE;
+}
+
+/* Returns the full path to the queried command */
+static const char *
+find_command (const char **known_locations)
+{
+	int i;
+
+	for (i = 0; known_locations [i]; i++){
+		if (g_file_test (known_locations [i], G_FILE_TEST_EXISTS))
+			return known_locations [i];
+	}
+	return NULL;
+}
+
+static void *
+unmount_thread_start (void *arg)
+{
+	UnmountData *data;
+	gint exit_status;
+	char *argv[5];
+	int i;
+	char *envp[] = {
+		"LC_ALL=C",
+		NULL
+	};
+
+	data = arg;
+
+	data->unmount_ok = TRUE;
+	
+	i = 0;
+	argv[i++] = (char *)data->command;
+	argv[i++] = data->device;
+	argv[i++] = NULL;
+	
+	if (g_spawn_sync (NULL,
+			  argv,
+			  envp,
+			  G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
+			  NULL, NULL,
+			  NULL,
+			  NULL,
+			  &exit_status,
+			  NULL)) {
+		if (exit_status == 0) {
+			data->unmount_ok = TRUE;
+		} else {
+			data->unmount_ok = FALSE;
+		}
+
+		/* Delay a bit to make sure unmount finishes */
+		sleep (2);
+	} else {
+		/* spawn failure */
+		data->unmount_ok = FALSE;
+	}
+
+	g_idle_add (unmount_done, data);	
+	
+	g_thread_exit (NULL); 	
+	
+	return NULL;
+}
+
+
+
+static gboolean
+unmount_drive (CDDrive *drive)
+{
+	UnmountData *data;
+	gboolean unmount_ok;
+	
+	if (drive->device == NULL)
+		return FALSE;
+	
+	unmount_ok = FALSE;
+
+	data = g_new0 (UnmountData, 1);
+	data->loop = g_main_loop_new (NULL, FALSE);
+	
+	data->timeout_tag = g_timeout_add (5*1000,
+					   unmount_timeout,
+					   data);
+	data->command = find_command (umount_known_locations);
+	data->device = g_strdup (drive->device);
+	g_thread_create (unmount_thread_start, data, FALSE, NULL);
+	
+	GDK_THREADS_LEAVE ();
+	g_main_loop_run (data->loop);
+	GDK_THREADS_ENTER ();
+	
+	g_main_loop_unref (data->loop);
+	data->loop = NULL;
+	
+	unmount_ok = data->unmount_ok;
+	
+	if (!data->timeout) {
+		/* Don't free data if mount operation still running. */
+		free_unmount_data (data);
+	}
+	
+	return unmount_ok;
+}
+
+
 static CDMediaType
 cd_recorder_wait_for_insertion (CDRecorder *cdrecorder, CDDrive *drive,
 				gboolean *needs_blank)
 {
 	CDMediaType type;
 	gboolean reload;
+	gboolean last_was_unmount;
 
 	reload = FALSE;
 	type = cd_drive_get_media_type_and_rewritable (drive, needs_blank);
@@ -500,17 +672,19 @@ cd_recorder_wait_for_insertion (CDRecorder *cdrecorder, CDDrive *drive,
 		reload = TRUE;
 	}
 
+	last_was_unmount = FALSE;
 	while (!media_type_matches (type)) {
 		gboolean res, busy_cd;
 
 		busy_cd = (type == CD_MEDIA_TYPE_BUSY);
 
-#ifdef USE_HAL
-		if (busy_cd != FALSE) {
-			/* TODO umount the disc if we find it,
-			 * and try it again */
+		if (busy_cd && !last_was_unmount) {
+			/* umount the disc if we find it, and try it again */
+			if (unmount_drive (drive)) {
+				last_was_unmount = TRUE;
+				goto retry;
+			}
 		}
-#endif
 
 		g_signal_emit (G_OBJECT (cdrecorder),
 				cd_recorder_table_signals[INSERT_CD_REQUEST],
@@ -521,6 +695,9 @@ cd_recorder_wait_for_insertion (CDRecorder *cdrecorder, CDDrive *drive,
 			return CD_MEDIA_TYPE_ERROR;
 		}
 
+		last_was_unmount = FALSE;
+
+	retry:
 		type = cd_drive_get_media_type_and_rewritable (drive, needs_blank);
 		reload = FALSE;
 		if (type == CD_MEDIA_TYPE_UNKNOWN
